@@ -22,11 +22,15 @@
 
 #define MAXLEN 0x400
 
+#define GZ_BUFFER_LEN 0x1000
+
 typedef struct nearest_module
 {
     /* The file to look at. */
     gzFile * gzfile;
     FILE * file;
+    unsigned char * gz_buffer;
+    int gz_buffer_at;
     /* Is this file a gz file? */
     int is_gz : 1;
     /* Print blah messages? */
@@ -82,6 +86,11 @@ static void nearest_compare_line (nearest_module_t * nearest)
 
     d = distance (b, l, nearest->search_term, nearest->search_len,
                   nearest->distance);
+    /*
+    printf ("%d ---> '%s' %d '%s' %d\n", d, 
+            nearest->search_term, nearest->search_len,
+            nearest->buf, l);
+    */
     if (d < nearest->distance) {
         if (nearest->verbose) {
             printf ("%s (%d) is nearer than %d.\n", nearest->buf, d,
@@ -95,6 +104,31 @@ static void nearest_compare_line (nearest_module_t * nearest)
     return;
 }
 
+static int
+nearest_gz_get_bytes (nearest_module_t * nearest)
+{
+    int ret;
+    /* Read some more bytes from the file. */
+    ret = gzread (nearest->gzfile, nearest->gz_buffer, GZ_BUFFER_LEN);
+    /* Error checking stuff. */
+    if (ret == Z_NULL) {
+        if (gzeof (nearest->gzfile)) {
+            return 0;
+        }
+        else {
+            int err;
+            const char * error_string;
+            error_string = gzerror (nearest->gzfile, & err);
+            fail (err, "Error reading %s: %s",
+                  nearest->file_name, error_string);
+        }
+    }
+    /* Set the pointer into the buffer back to the start of the
+       buffer. */
+    nearest->gz_buffer_at = 0;
+    return 1;
+}
+
 static void
 nearest_open_file (nearest_module_t * nearest)
 {
@@ -115,6 +149,13 @@ nearest_open_file (nearest_module_t * nearest)
           (!nearest->is_gz && !nearest->file),
           "Cannot open package file %s: %s",
           nearest->file_name, strerror (errno));
+    if (nearest->is_gz) {
+        nearest->gz_buffer = malloc (GZ_BUFFER_LEN);
+        fail (! nearest->gz_buffer, "no memory");
+        if (! nearest_gz_get_bytes (nearest)) {
+            fail (1, "File is empty");
+        }
+    }
     return;
 }
 
@@ -130,33 +171,62 @@ nearest_close_file (nearest_module_t * nearest)
     return;
 }
 
+static int
+nearest_gz_get_line (nearest_module_t * nearest)
+{
+    /* The number of bytes remaining in "nearest->gz_buffer". */
+    int remaining;
+    /* Boolean, set to true if we need to read more bytes from the
+       gzfile. */
+    int read_more_bytes;
+    /* The number of bytes read into "nearest->buf so far". */
+    int bt;
+    int i;
+
+    read_more_bytes = 1;
+    /* Initially, "bt" is zero, but if more bytes need to be read from
+       the .gz file, it can be incremented. */
+    bt = 0;
+ more_bytes:
+
+    remaining = GZ_BUFFER_LEN - nearest->gz_buffer_at;
+    for (i = 0; i < remaining; i++) {
+        nearest->buf[bt] = nearest->gz_buffer[nearest->gz_buffer_at];
+        /* If there is a carriage return in the stuff read so far, we
+           don't need to read any more bytes, so we can just leave
+           this routine now, without calling the gzlib. */
+        nearest->gz_buffer_at++;
+        if (nearest->buf[bt] == '\n') {
+            read_more_bytes = 0;
+            nearest->buf[bt] = '\0';
+            break;
+        }
+        bt++;
+    }
+    if (read_more_bytes) {
+        if (! nearest_gz_get_bytes (nearest)) {
+            return 0;
+        }
+        goto more_bytes;
+    }
+    return 1;
+}
+
 static int
 nearest_get_line (nearest_module_t * nearest)
 {
     char * ret;
 
     if (nearest->is_gz) {
-        ret = gzgets (nearest->gzfile, nearest->buf, MAXLEN);
-        if (ret == Z_NULL) {
-            if (! gzeof (nearest->gzfile)) {
-                int err;
-                const char * error_string;
-                error_string = gzerror (nearest->gzfile, & err);
-                fail (err, "Error reading %s: %s",
-                      nearest->file_name, error_string);
-            }
-            return 0;
-        }
+        return nearest_gz_get_line (nearest);
     }
-    else {
-        ret = fgets (nearest->buf, MAXLEN - 1, nearest->file);
-        if (! ret) {
-            if (! feof (nearest->file)) {
-                fail (1, "Error reading %s: %s",
-                      nearest->file_name, strerror (errno));
-            }
-            return 0;
+    ret = fgets (nearest->buf, MAXLEN - 1, nearest->file);
+    if (! ret) {
+        if (! feof (nearest->file)) {
+            fail (1, "Error reading %s: %s",
+                  nearest->file_name, strerror (errno));
         }
+        return 0;
     }
     return 1;
 }
@@ -164,13 +234,16 @@ nearest_get_line (nearest_module_t * nearest)
 static void search_packages (nearest_module_t * nearest)
 {
     static int max_sane_distance = 10;
+    static int more_lines;
     nearest_open_file (nearest);
     /* Don't use INT_MAX here or get overflow. */
     nearest->distance = nearest->search_len + 1;
     if (nearest->search_len > max_sane_distance) {
         nearest->distance = max_sane_distance;
     }
-    while (nearest_get_line (nearest)) {
+    more_lines = 1;
+    while (more_lines) {
+        more_lines = nearest_get_line (nearest);
         nearest_compare_line (nearest);
     }
     nearest_close_file (nearest);
@@ -247,7 +320,7 @@ int main (int argc, char ** argv)
         st = "Lingua::Stop::Weirds";
     }
     nearest_set_search_term (& nearest, st);
-    nearest_set_search_file (& nearest, "02packages.details.txt");
+    nearest_set_search_file (& nearest, "02packages.details.txt.gz");
 
     search_packages (& nearest);
     print_result (& nearest);
